// =====================================================================
//
// XBUS Servo Programmer Shield
//
// 2022/09/19
// by ZakLab / Zak Sawa
//
// This code is tested on Arduino UNO R3, M0
//
// =====================================================================

#include <I2CLiquidCrystal.h>
#include <XBusServo.h>


#if defined( __SAMD21G18A__ )
#define kUseMsTimer2    0       // use polling
#else
#define kUseMsTimer2    1       // use interrupt
#endif

#define kLCD5V          true    // true if 5V LCD



#if kUseMsTimer2
#include <MsTimer2.h>
#endif




//
// defines
//
#define kMaxServoNum      	16		// 1 - 50
#define kDirPinNum        	2		// pin number for direction
#define kRotaryInputA     	3		// pin number to input rotary phase A (INT1)
#define kRotaryInputB     	4		// pin number to input rotary phase B
#define kUpButton         	5		// pin number for UP button
#define kDownButton       	6		// pin number for DOWN button
#define kRightButton      	7		// pin number for RIGHT button
#define kLeftButton			8		// pin number for LEFT button

#define kMaxServoCommand	(sizeof(commandDataBase) / sizeof(servoCommandRec))
#define	CurrentID			((gCurrentID & 0x3F) | (((gCurrentSubID - 1) << 6) & 0xC0))

#define kGainMax          	50

#define kCursorBlink      	500     // mSec
#define kCursorChar       	0x07
#define kLCD_Row          	16
#define kLCD_Column       	2
#define kValueDigit       	4



//
// typedefs
//
typedef enum
{
  kCursor_ID,
  kCursor_SubID,
  kCursor_Function,
  kCursor_Param
}
cursorType;

typedef struct servoCommandRec
{
  char                        name[13];
  XBusOrder                   order;
  uint8_t                     writeIndex;
  uint8_t                     hexData;        // true if the data should be show in Hex
  uint8_t                     unsignedData;   // true if the data is unsigned
  int32_t                     maxValue;
  int32_t                     minValue;
  int32_t                     step;           // another step when fast key is pressed
}
servoCommandRec;

typedef enum
{
  kRotaly_Stop,
  kRotaly_Right,
  kRotaly_Left
}
rotalyType;




//
// servo command data base
//

PROGMEM static const servoCommandRec    commandDataBase[] = {
  //  name          order                           write index                 hex   unsign  max       min         step
  {"Position   :",  (XBusOrder)0x00,                0,                          1,    1,      0x0FFFF,  0x0000,     300},
  {"Servo ID   :",  (XBusOrder)0x00,                0,                          0,    0,      50,       1,          1},
  {"SubID      :",  (XBusOrder)0x00,                0,                          0,    0,      4,        1,          1},
  {"Version    :",  kXBusOrder_Version,             0,                          1,    1,      0,        0,          0},
  {"ModelNumber:",  kXBusOrder_Product,             0,                          1,    1,      0,        0,          0},
  {"Reverse    :",  kXBusOrder_Reverse,             kParamIdx_Reversed,         0,    0,      1,        0,          1},
  {"Neutral    :",  kXBusOrder_Neutral,             kParamIdx_NeutralOffset,    0,    0,      600,      -600,       10},
  {"Travel High:",  kXBusOrder_H_Travel,            kParamIdx_TravelHigh,       0,    0,      192,      0,          5},
  {"Travel Low :",  kXBusOrder_L_Travel,            kParamIdx_TravelLow,        0,    0,      192,      0,          5},
  {"Limit High :",  kXBusOrder_H_Limit,             kParamIdx_LimitHigh,        1,    1,      0x0FFFF,  0x0000,     10},
  {"Limit Low  :",  kXBusOrder_L_Limit,             kParamIdx_LimitLow,         1,    1,      0x0FFFF,  0x0000,     10},
  {"Holding (P):",  kXBusOrder_P_Gain,              kParamIdx_PGainDiff,        0,    0,      kGainMax, -kGainMax,  1},
  {"Integer (I):",  kXBusOrder_I_Gain,              kParamIdx_IGainDiff,        0,    0,      kGainMax, -kGainMax,  1},
  {"Damping (D):",  kXBusOrder_D_Gain,              kParamIdx_DGainDiff,        0,    0,      kGainMax, -kGainMax,  1},
  {"Integer Max:",  kXBusOrder_MaxInteger,          kParamIdx_MaxIntegerDiff,   0,    0,      999,      -999,       10},
  {"Dead Band  :",  kXBusOrder_DeadBand,            kParamIdx_DeadBandDiff,     0,    0,      40,       -20,        1},
  {"Angle Mode :",  kXBusOrder_AngleMode,           kParamIdx_AngleMode,        0,    0,      (kXBusAngleMode_MaxNum - 1),    -1,     1},
  {"Speed Limit:",  kXBusOrder_SpeedLimit,          kParamIdx_SpeedLimit,       0,    0,      30,       0,          1},
  {"Stop Mode  :",  kXBusOrder_StopMode,            kParamIdx_StopMode,         0,    0,      1,        0,          1},
  {"Boost      :",  kXBusOrder_PowerOffset,         kParamIdx_PWOffsetDiff,     0,    0,      999,      -999,       10},
  {"Slow Start :",  kXBusOrder_SlowStart,           kParamIdx_SlowStart,        0,    0,      1,        0,          1},
  {"PWM Mode   :",  kXBusOrder_PWMMode,             kParamIdx_PWMMode,          0,    0,      (kXBusPWMMode_MaxNum - 1),      0,      1},
  {"Interp Mode:",  kXBusOrder_InterpolateMode,     kParamIdx_InterpolateMode,  0,    0,      1,        0,          1},
  {"Alarm Level:",  kXBusOrder_AlarmLevel,          kParamIdx_AlarmLevel,       0,    0,      99,       0,          10},
  {"Alarm Delay:",  kXBusOrder_AlarmDelay,          kParamIdx_AlarmDelay,       0,    0,      5000,     0,          100},
  {"CurPosition:",  kXBusOrder_CurrentPos,          0,                          1,    0,      0,        0,          0},
  {"Cur Power  :",  kXBusOrder_CurrentPow,          0,                          0,    0,      0,        0,          0},
};


//
// global vars
//
XBusServo			        gXBusServo(kDirPinNum, kMaxServoNum);
bool	    	          	gOK2SendPacket = true;
bool	    		        gSaveCurrentValue = false;
bool		    	        gCurrentValueChanged = false;
bool		    	        gDirty = true;			                  // true for first update
long int              		gCurrentPos = kXbusServoNeutral;
long int              		gCurrentValue = 0;

I2CLiquidCrystal	    	gLCD(40, kLCD5V);
cursorType			      	gCurrentCursor = kCursor_ID;
bool                  		gCursorON = false;

volatile rotalyType   		gRotalyJob = kRotaly_Stop;
unsigned char		    	gCurrentFunction = 0;
servoCommandRec       		gCommandData;
unsigned char		    	gNextFunction = 0;
unsigned char		    	gCurrentID = 1;
unsigned char		    	gCurrentSubID = 1;
bool                  		gNotSupported = false;



//=============================================================
// sendPacket()
//  This is the handler to keep to send channel command packet with 14mSec interval.
//=============================================================
void sendPacket()
{
  if (gOK2SendPacket)
    gXBusServo.sendChannelDataPacket();
}


//=============================================================
// setup()
//	default function for Arduino
//=============================================================
void setup()
{
  static const uint8_t   CursorCharMap[] = {0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF};

  // XBus setup
  gXBusServo.begin();
  gXBusServo.addServo(0x01, kXbusServoNeutral);		 // add first servo with channelID = 0x01
#if kUseMsTimer2
  MsTimer2::set(kXBusInterval, sendPacket);
  MsTimer2::start();
#endif

  // LCD setup
  gLCD.begin(kLCD_Row, kLCD_Column);
  gLCD.createChar(kCursorChar, (uint8_t*)CursorCharMap);

  // Rotary encorder setup
  pinMode(kRotaryInputA, INPUT);
  pinMode(kRotaryInputB, INPUT);
  digitalWrite(kRotaryInputA, HIGH);
  digitalWrite(kRotaryInputB, HIGH);
  attachInterrupt(digitalPinToInterrupt(3), getRotary, CHANGE);

  // Buttons setup
  pinMode(kUpButton, INPUT);
  pinMode(kDownButton, INPUT);
  pinMode(kRightButton, INPUT);
  pinMode(kLeftButton, INPUT);
  digitalWrite(kUpButton, HIGH);
  digitalWrite(kDownButton, HIGH);
  digitalWrite(kRightButton, HIGH);
  digitalWrite(kLeftButton, HIGH);

  // data setup
  memcpy_P (&gCommandData, &(commandDataBase[0]), sizeof(gCommandData));
}


//=============================================================
// rotaryRight()
//	This is what to do when rotary encoder turn right
//=============================================================
void rotaryRight()
{
  if (gDirty)
    return;

  switch (gCurrentCursor)
  {
    case kCursor_ID:
      if (gCurrentID < kXBusMaxServoNum)
      {
        gXBusServo.removeServo(CurrentID);
        gCurrentID++;
        gXBusServo.addServo(CurrentID, gCurrentPos);
        gDirty = true;
      }
      break;

    case kCursor_SubID:
      if (gCurrentSubID < gCommandData.maxValue)
      {
        gXBusServo.removeServo(CurrentID);
        gCurrentSubID++;
        gXBusServo.addServo(CurrentID, gCurrentPos);
        gDirty = true;
      }
      break;

    case kCursor_Function:
      if (gCurrentFunction < (kMaxServoCommand - 1))
      {
        gNextFunction = gCurrentFunction + 1;
        gDirty = true;
      }
      break;

    case kCursor_Param:
      if (gCommandData.maxValue == gCommandData.minValue)
        break;

      if (gCurrentFunction == 0)
      {
        if (digitalRead(kRightButton))
          gCurrentPos += 10;
        else
          gCurrentPos += gCommandData.step;

        if (gCurrentPos >= gCommandData.maxValue)
          gCurrentPos = gCommandData.maxValue;
        gDirty = true;
      }
      else if (gCurrentValue < gCommandData.maxValue)
      {
        if (digitalRead(kRightButton))
          gCurrentValue++;
        else
          gCurrentValue += gCommandData.step;

        if (gCurrentValue >= gCommandData.maxValue)
          gCurrentValue = gCommandData.maxValue;
        gCurrentValueChanged = true;
        gDirty = true;
      }
      break;
  }
}


//=============================================================
// rotaryLeft()
//	This is what to do when rotary encoder turn left
//=============================================================
void rotaryLeft()
{
  if (gDirty)
    return;

  switch (gCurrentCursor)
  {
    case kCursor_ID:
      if (gCurrentID > 1)
      {
        gXBusServo.removeServo(CurrentID);		 // add first servo with channelID = 0x01
        gCurrentID--;
        gXBusServo.addServo(CurrentID, gCurrentPos);		 // add first servo with channelID = 0x01
        gDirty = true;
      }
      break;

    case kCursor_SubID:
      if (gCurrentSubID > gCommandData.minValue)
      {
        gXBusServo.removeServo(CurrentID);		 // add first servo with channelID = 0x01
        gCurrentSubID--;
        gXBusServo.addServo(CurrentID, gCurrentPos);		 // add first servo with channelID = 0x01
        gDirty = true;
      }
      break;

    case kCursor_Function:
      if (gCurrentFunction > 0)
      {
        gNextFunction = gCurrentFunction - 1;
        gDirty = true;
      }
      break;

    case kCursor_Param:
      if (gCommandData.maxValue == gCommandData.minValue)
        break;

      if (gCurrentFunction == 0)
      {
        if (digitalRead(kRightButton))
          gCurrentPos -= 10;
        else
          gCurrentPos -= gCommandData.step;

        if (gCurrentPos <= gCommandData.minValue)
          gCurrentPos = gCommandData.minValue;
        gDirty = true;
      }
      else if (gCurrentValue > gCommandData.minValue)
      {
        if (digitalRead(kRightButton))
          gCurrentValue--;
        else
          gCurrentValue -= gCommandData.step;

        if (gCurrentValue <= gCommandData.minValue)
          gCurrentValue = gCommandData.minValue;
        gCurrentValueChanged = true;
        gDirty = true;
      }
      break;
  }
}


//=============================================================
// getRotary()
//	This is the handler to get rotary encoder
//=============================================================
void getRotary()
{
  static unsigned char sOldRot = 0;

  if (! digitalRead(kRotaryInputA))
  { // Check to start rotating
    if (digitalRead(kRotaryInputB))
      sOldRot = 'R'; // Right turn started
    else
      sOldRot = 'L'; // Left turn started
  }
  else
  { // Check to stop rotating
    if (digitalRead(kRotaryInputB))
    {
      if (sOldRot == 'L') // It's still left turn
        if (gRotalyJob == kRotaly_Stop)
          gRotalyJob = kRotaly_Left;
    }
    else
    {
      if (sOldRot == 'R')	 // It's still right turn
        if (gRotalyJob == kRotaly_Stop)
          gRotalyJob = kRotaly_Right;
    }

    sOldRot = 0;
  }
}


//============================================================================
// HexString
//  Convert to HEX String
//============================================================================

void HexString(char*                    theStr,
               uint32_t                  theValue,
               uint32_t                  digit)
{

  uint32_t            counter;
  uint8_t             temp;

  for (counter = 0; counter < digit; counter++)
  {
    temp = theValue >> ((digit - counter - 1) * 4);
    temp &= 0x0F;

    if (temp <= 9)
      theStr[counter] = temp + 0x30;
    else
      theStr[counter] = temp - 9 + 0x40;
  }
}


//============================================================================
// DecString
//  Convert to Dec String
//============================================================================

void DecString(char*                    theStr,
               int32_t                   theValue,
               uint32_t                  digit,
               uint8_t                   zeroSupress)
{

  uint32_t            counter;
  int32_t             base = 1;
  int32_t             temp;
  uint8_t             minusValue = 0;

  for (counter = 1; counter < digit; counter++)
    base *= 10;

  minusValue = (theValue < 0);
  if (minusValue)
    theValue = -theValue;

  for (counter = 0; counter < digit; counter++)
  {
    temp = theValue / base;
    if ((temp == 0) & zeroSupress)
    {
      if (base == 1)
        theStr[counter] = '0';
      else
        theStr[counter] = ' ';
    }
    else
    {
      zeroSupress = 0;

      if (minusValue & (counter > 0))
      {
        theStr[counter - 1] = '-';
        minusValue = 0;
      }

      theStr[counter] = temp + 0x30;
    }

    theValue -= temp * base;
    base /= 10;
  }
}


//=============================================================
// updateLCD()
//	update contents of LCD
//  Why not using String ? Because it's toooooo FAT
//=============================================================
void updateLCD()
{
  char                              lineBuffer[kLCD_Row + 1] = "ID00-0          ";
  PROGMEM static const char         angle120[] =            " 120";
  PROGMEM static const char         angle180[] =            " 180";
  PROGMEM static const char         angle150[] =            " 150";
  PROGMEM static const char         angleMax[] =            " Max";
  PROGMEM static const char         holdHold[] =            "Hold";
  PROGMEM static const char         holdFree[] =            "Free";
  PROGMEM static const char         boolOn[] =              "  on";
  PROGMEM static const char         boolOff[] =             " off";
  PROGMEM static const char         pwm_jr_s[] =            "JR S";
  PROGMEM static const char         pwm_fu_s[] =            "Fu S";
  PROGMEM static const char         pwm_fu_n[] =            "Fu N";
  PROGMEM static const char         pwm_sa_n[] =            "Sa N";
  PROGMEM static const char         notSupported[] =        "----";

  // update ID
  DecString(&lineBuffer[2], gCurrentID, 2, 0);
  DecString(&lineBuffer[5], gCurrentSubID, 1, 0);

  // update cursor for 1st line
  if (gCursorON)
    switch (gCurrentCursor)
    {
      case kCursor_ID:
        lineBuffer[3] = kCursorChar;
        break;

      case kCursor_SubID:
        lineBuffer[5] = kCursorChar;
        break;
    }

  gLCD.setCursor(0, 0);
  gLCD.print(lineBuffer);

  // update function name
  memcpy(lineBuffer, gCommandData.name, (kLCD_Row - kValueDigit));

  // update current value
  switch (gCurrentFunction)
  {
    case 0:     // position
      HexString(&lineBuffer[12], gCurrentPos, kValueDigit);
      break;

    case 1:     // ID
    case 2:     // Sub ID
      DecString(&lineBuffer[12], gCurrentValue, kValueDigit, 1);
      break;

    default:
      if (gNotSupported)
      {
        memcpy_P(&lineBuffer[12], notSupported, kValueDigit);
        break;
      }

      switch (gCommandData.order)
      {
        case kXBusOrder_AngleMode:
          switch (gCurrentValue)
          {
            case kXBusAngleMode_Max:
              memcpy_P(&lineBuffer[12], angleMax, kValueDigit);
              break;

            case kXBusAngleMode_120:
            default:
              memcpy_P(&lineBuffer[12], angle120, kValueDigit);
              break;

            case kXBusAngleMode_180:
              memcpy_P(&lineBuffer[12], angle180, kValueDigit);
              break;

            case kXBusAngleMode_150:
              memcpy_P(&lineBuffer[12], angle150, kValueDigit);
              break;
          }
          break;

        case kXBusOrder_PWMMode:
          switch (gCurrentValue)
          {
            case kXBusPWMMode_Normal:
            default:
              memcpy_P(&lineBuffer[12], pwm_jr_s, kValueDigit);
              break;

            case kXBusPWMMode_F_Normal:
              memcpy_P(&lineBuffer[12], pwm_fu_s, kValueDigit);
              break;

            case kXBusPWMMode_F_Narrow:
              memcpy_P(&lineBuffer[12], pwm_fu_n, kValueDigit);
              break;

            case kXBusPWMMode_S_SR:
              memcpy_P(&lineBuffer[12], pwm_sa_n, kValueDigit);
              break;
          }
          break;

        case kXBusOrder_StopMode:
          switch (gCurrentValue)
          {
            case 0:
            default:
              memcpy_P(&lineBuffer[12], holdFree, kValueDigit);
              break;

            case 1:
              memcpy_P(&lineBuffer[12], holdHold, kValueDigit);
              break;
          }
          break;

        default:
          if (gCommandData.hexData)
            HexString(&lineBuffer[12], gCurrentValue, kValueDigit);
          else if ((gCommandData.maxValue == 1) && (gCommandData.minValue == 0))
          {
            if (gCurrentValue == 1)
              memcpy_P(&lineBuffer[12], boolOn, kValueDigit);
            else
              memcpy_P(&lineBuffer[12], boolOff, kValueDigit);
          }
          else
            DecString(&lineBuffer[12], gCurrentValue, kValueDigit, 1);
          break;
      }
  }

  // update cursor for 2nd line
  if (gCursorON)
    switch (gCurrentCursor)
    {
      case kCursor_Function:
        lineBuffer[0] = kCursorChar;
        break;

      case kCursor_Param:
        lineBuffer[15] = kCursorChar;
        break;
    }

  gLCD.setCursor(0, 1);
  gLCD.print(lineBuffer);
}


//=============================================================
// buttonHandler()
//	get push button status and change flags
//=============================================================
void buttonHandler()
{
  if (! digitalRead(kUpButton))
  {
    if (gCurrentCursor == kCursor_Function)
    {
      gCurrentCursor = kCursor_ID;
      gDirty = true;
    }
    else if (gCurrentCursor == kCursor_Param)
    {
      gCurrentCursor = kCursor_SubID;
      gDirty = true;
      gSaveCurrentValue = true;
    }
  }
  else if (! digitalRead(kDownButton))
  {
    if (gCurrentCursor == kCursor_ID)
    {
      gCurrentCursor = kCursor_Function;
      gDirty = true;
    }
    else if (gCurrentCursor == kCursor_SubID)
    {
      gCurrentCursor = kCursor_Param;
      gDirty = true;
    }
  }
  else if (! digitalRead(kRightButton))
  {
    if (gCurrentCursor == kCursor_ID)
    {
      gCurrentCursor = kCursor_SubID;
      gDirty = true;
    }
    else if (gCurrentCursor == kCursor_Function)
    {
      gCurrentCursor = kCursor_Param;
      gDirty = true;
    }
  }
  else if (! digitalRead(kLeftButton))
  {
    if (gCurrentCursor == kCursor_SubID)
    {
      gCurrentCursor = kCursor_ID;
      gDirty = true;
    }
    else if (gCurrentCursor == kCursor_Param)
    {
      gCurrentCursor = kCursor_Function;
      gDirty = true;
      gSaveCurrentValue = true;
    }
  }
}


//=============================================================
// setCurrentValue()
//	set current value to the servo for temp
//=============================================================
void setCurrentValue()
{
  int					theValue;
  XBusError		result;

  // set current value
  if (gCurrentValueChanged)
  {
    if (gCommandData.writeIndex != 0)
    {
      theValue = gCurrentValue;
      for (;;)
      {
        result = gXBusServo.setCommand(CurrentID, gCommandData.order, &theValue);
        delay(10);
        if (result == kXBusError_NoError)
          break;
      }
    }
  }
}


//=============================================================
// writeCurrentValue()
//	write current value to the servo
//=============================================================
void writeCurrentValue()
{
  XBusError		result;

  if (gCommandData.writeIndex == 0)
  {
    // for change ID
    if ((gCurrentFunction == 1) || (gCurrentFunction == 2))
    {
      int			newChannelID;

      if (gCurrentFunction == 1)
        newChannelID = (gCurrentValue & 0x3F) | (((gCurrentSubID - 1) << 6) & 0xC0);
      else
        newChannelID = (gCurrentID & 0x3F) | (((gCurrentValue - 1) << 6) & 0xC0);

      gXBusServo.removeServo(CurrentID);

      for (;;)
      {
        result = gXBusServo.setChannelID(CurrentID, newChannelID);
        delay(10);
        if (result == kXBusError_NoError)
          break;
      }

      if (gCurrentFunction == 1)
        gCurrentID = gCurrentValue;
      else
        gCurrentSubID = gCurrentValue;

      gXBusServo.addServo(CurrentID, gCurrentPos);
    }
  }
  else
  {
    int			writeIndex;

    writeIndex = gCommandData.writeIndex;
    for (;;)
    {
      result = gXBusServo.setCommand(CurrentID, kXBusOrder_2_ParamWrite, &writeIndex);
      delay(10);
      if (result == kXBusError_NoError)
        break;
    }
  }
}


//=============================================================
// getCurrentValue()
//	get current value from the servo
//=============================================================
void getCurrentValue()
{
  int					theValue;
  XBusError		result;

  // get current value
  gNotSupported = false;
  switch (gCurrentFunction)
  {
    case 0:
    case 1:
    case 2:
      // do nothing
      break;

    default:
      for (;;)
      {
        result = gXBusServo.getCommand(CurrentID, gCommandData.order, &theValue);
        delay(10);
        if (result == kXBusError_NoError)
          break;

        if (result == kXBusError_TimeOut)
        {
          gNotSupported = true;
          break;
        }
      }
      gCurrentValue = theValue;
      if (gCommandData.unsignedData)
        gCurrentValue &= 0x0000FFFF;
      break;
  }
}


//=============================================================
// loop()
//	default function for Arduino
//=============================================================
void loop()
{
  static uint32_t           sCursorCounter = 0;
#if !kUseMsTimer2
  static uint32_t           sXBUSCounter = 0;
#endif

  buttonHandler();

  switch (gRotalyJob)
  {
    case kRotaly_Right:
      rotaryRight();
      break;

    case kRotaly_Left:
      rotaryLeft();
      break;

    case kRotaly_Stop:
    default:
      break;			// Do nothing
  }

  // parameter control
  if (gDirty)
  {
    gOK2SendPacket = false;

    if (! gNotSupported)
    {
      setCurrentValue();

      if (gSaveCurrentValue)
        writeCurrentValue();
    }

    if (gNextFunction != gCurrentFunction)
    {
      gCurrentFunction = gNextFunction;
      memcpy_P (&gCommandData, &(commandDataBase[gCurrentFunction]), sizeof(gCommandData));

      if (gCurrentFunction == 1)
        gCurrentValue = gCurrentID;
      else if (gCurrentFunction == 2)
        gCurrentValue = gCurrentSubID;
      else
        getCurrentValue();
    }

    updateLCD();

    gXBusServo.setServo(CurrentID, gCurrentPos);

    gOK2SendPacket = true;
  }

  // update cursor blink
  if (millis() >= sCursorCounter)
  {
    sCursorCounter = millis() + kCursorBlink;
    gCursorON = ! gCursorON;
    updateLCD();
  }

#if !kUseMsTimer2
  // use polling
  if (millis() >= sXBUSCounter)
  {
    if (sXBUSCounter == 0)
      sXBUSCounter = millis() + kXBusInterval;  // for init time loss
    else
      sXBUSCounter += kXBusInterval;

    sendPacket();
  }
#endif

  gDirty = false;
  gSaveCurrentValue = false;
  gRotalyJob = kRotaly_Stop;
}
